/*
   \TEX{
   \Section{Big integers libraries}
   \Subsection{Big integers layer of \zen}
   The |ZEN| library was primarily built upon the |BigNum| library~\cite{bignum}. Since then,
   we have extended |ZEN| so that it can now use |GMP|,
   the Gnu Multiprecision Package~\cite{gmp}.
   Our final purpose is to be able to use the most well suited function
   where we need it in the library. For the present version,
   this is not the case. The internal call in |ZEN|'s function are
   still in the |BigNum|'s format, and using |GMP| means that the |BigNum|'s
   function are emulated. However, we have already noticed some improvements
   for a certain number of architectures.

   In order to achieve this final goal, all the big integers functions are called in |ZEN| through a reformatted version of both |BigNum| and |GMP|' functions. These macros are prefixed by |ZBN|.
   Besides, some additional functionalities were needed in upper layers of |ZEN|.

   \begin{note}
   {\bf The |ZBN| format is entirely made of macros. It is therefore impossible to use side effects in parameters of |ZBN| calls.}
   \end{note}
   \input{zbngmp-imakefile}

   } 
   */

#ifndef __ZBN__
#define __ZBN__

/*
   \TEX{
   \Section{The |ZBN| layer}
   We now describe the |ZBN| layer of \zen.
   \Subsection{Parameters}
   \Subsubsection{Memory limitation}
   \label{zbncompil}

   Some of the new functionalities added in the |ZBN| layer, need a lot of memory at compilation time, specially with the |gcc| compiler. You may decrease the amount of memory needed by using the following compilation flag which is set in |zbn.h|.
   } 
   */




#ifndef LOW_MEMORY		/* If set do not inline datas
				   in zbn_weight_data.h
				   \CiteC */
# define LOW_MEMORY 0		/* \EndCiteC */
#endif

/*
   \TEX{
   \Subsubsection{Karatsuba's multiplication on integers}
The Karatsuba's multiplication on big integers is used by default in ZEN. This can be disabled using the |KARA| parameter.
   } 
   */
#ifndef KARA			/* Activate Karatsuba operations
				   on integers \CiteC */
# define KARA 1			/* \EndCiteC */
#endif

#include <stdio.h>
#include <math.h>
#include "sys.h"

#define LN2PREZENMULT 8		/* Used in precomputations in GF(2)
				   value of 8 indicates tabulation on
				   8 bits. */

#define false 0
#define true 1

#define ZBN_HAS_INVERSE 0
#define ZBN_NO_INVERSE  1
#define FMT_BINARY 2

/* \TEX{
   \Subsection{Constants}
   Three constants are used in |ZEN|:
   \begin{description}
   \item[|SIZE_BLOC|] depends on the arithmetic of
   computers. It is the size in bits of a |BigNumDigit|.
   \item[|SIZE_CHAR|] is the size in bits of a |char| (usually 8 bits).
   \item[|SIZE_SHORT|] is the size in bits of a |short| (usually 16 bits).
   \end{description}
   }
   */
#ifdef HasBigNum
# include "BigNum.h"

# define SIZE_BLOC BN_DIGIT_SIZE  
# define SIZE_SHORT sizeof(short)*BN_BYTE_SIZE  
# define SIZE_CHAR BN_BYTE_SIZE  
#else  /* GMP */
# include "gmp.h"
# ifdef sun4
  /* sordid hacks to avoid conflicts with stdlib.h on old sparc computers */
#  define malloc temporarygmpmalloc
#  define realloc temporarygmprealloc
#  define free temporarygmpfree
# endif
# include "gmp-impl.h"
# ifdef sun4
  /* sordid hacks to avoid conflicts with stdlib.h on old sparc computers */
#  undef malloc
#  undef realloc
#  undef free
# endif

# define SIZE_BLOC BITS_PER_MP_LIMB
# define SIZE_SHORT BITS_PER_SHORTINT
# define SIZE_CHAR BITS_PER_CHAR

typedef  mp_limb_t BigNumDigit;
typedef  mp_limb_t *BigNum;
typedef  mp_size_t BigNumLength;

/* FOR COMPATIBLITY ONLY, it should disappear.. */
# define BN_DIGIT_SIZE BITS_PER_MP_LIMB
# define BN_BYTE_SIZE BITS_PER_CHAR

# define BN_LT -1
# define BN_EQ 0
# define BN_GT 1
# ifndef SQUARING_BOUND
#  define SQUARING_BOUND 8
# endif

# ifndef TRUE
#  define TRUE 1
# endif
# ifndef FALSE
#  define FALSE 0
# endif
#endif /* BigNum/GMP */


/**********************************************************************/
/*
   \TEX{
   \Subsection{Basic functions on |BigNumDigit|s}
   We defined the following macros to help manipulating a  |BigNum|.
   These can also be used under GMP compilation.
   In fact, these macros should always be used when manipulating big integers.
   }
   */
     
#if 0
/*
   \TEX{
   \title{Reduction modulo a bloc}
   \input{An integer |x|.}
   \output{The remainder of $|x| / |SIZE_BLOC|$, that is to say $|x| \bmod |SIZE_BLOC|$ }
   }
   */
BigNumLength modSizeBloc(x)
   BigNumLength x;{}
#endif
#define modSizeBloc(x) ((x) % SIZE_BLOC)
#if 0
/*
   \TEX{
   \title{Number of blocs}
   \input{An integer |x|.}
   \output{The quotient $|x| / |SIZE_BLOC|$ }
   }
   */
BigNumLength divSizeBloc(x)
   BigNumLength x;{}
#endif
#define divSizeBloc(x) ((x) / SIZE_BLOC)
#if 0
/*
   \TEX[
   The two following functions have an equivalent meaning, once replaced |SIZE_BLOC| by |SIZE_CHAR|.
   \title{Reduction modulo a char}
   \input{An integer}
   \output{The remainder of $|x| / |SIZE_CHAR|$, that is to say $|x| \bmod |SIZE_CHAR|$ }
   }
   */
BigNumLength modSizeChar(x)
   BigNumLength x;{}
#endif
#define modSizeChar(x) ((x) % SIZE_CHAR)
#if 0
/*
   \TEX[
   \title{Number of chars}
   \input{An integer}
   \output{The quotient $|x| / |SIZE_CHAR|$ }
   }
   */
BigNumLength divSizeChar(x)
   BigNumLength x;{}
#endif
#define divSizeChar(x) ((x) / SIZE_CHAR)

#if 0
/*
   \TEX{
   \title{Exponentiation}
   \input{An integer |x|.}
   \output{The |BigNumDigit| representing $2^{|x|}$ if $|x|<|SIZE_BLOC|$, 0 otherwise. }
   } 
   */
BigNumDigit BlocExp2(x)
     unsigned int x; {}
#endif

#if 1
# define BlocExp2(x) (((unsigned int)(x))>=SIZE_BLOC ?(BigNumDigit)(0):((BigNumDigit)1<<(x)))
#else
extern const BigNumDigit ___Exp2[64];
# define BlocExp2(x) (((unsigned int)(x))>=SIZE_BLOC ? (BigNumDigit)0 : ___Exp2[x])
#endif
#if 0
/*
   \TEX{
   \title{Getting least significant digit of a |BigNum|}
   \input{An allocated |BigNum|.}
   \output{The least significant digit of |n|.}
   } 
   */
BigNumDigit ZBNGetDigit(n)
     BigNum n; {}

/*
   \TEX{
   \title{Assigning least significant digit of a |BigNum|}
   \input{An allocated |BigNum| and a |BigNumDigit|.}
   \side{The least significant digit of |n| is set to |d|.}
   } 
   */
void ZBNSetDigit(n,d)
     BigNum n;
     BigNumDigit d;{}

/*
\TEX{
   \title{Getting i$^{\rm th}$ digit of a |BigNum|}
   \input{An integer and an allocated |BigNum|.}
   \output{The |i|th digit of |n|.}
   \note{The 0th digit is the least significant one. If |n| is of size |nl|, the (|nl|-1)th is the most significant one.
   This is a macro that can be a lvalue in an assignment.}
   } 
   */
BigNumDigit ZBNDigit(n,i)
     BigNum n;
     int i; {}
#endif
#define ZBNDigit(n,i) ((n)[i])
#ifdef HasBigNum
# include "zbn-bn.h"
#else
# include "zbn-gmp.h"
#endif

#if 0
/*
   \TEX{
   \title{Number of unsignificant bits}
   \input{A digit |d|.}
   \output{The number |k| of most significant bits set to zero in |d|.}
   \note{If |d| is zero, $|k|=|SIZE_BLOC|$. }
   } 
   */
unsigned int ZBNNumLeadingZeroBitsInDigit(d)
     BigNumDigit d; {}

/*
   \TEX{
   \title{Number of significative bits}
   \input{A digit |d|.}
   \output{The number |k| of significative bits in |d|.}
   \note{If |d| is zero, $|k|=0$. }
   } 
   */
unsigned int ZBNNumBitsInDigit(d)
     BigNumDigit d; {}
#endif
#define ZBNNumBitsInDigit(d) (SIZE_BLOC-ZBNNumLeadingZeroBitsInDigit(d))

#if 0
/*
   \TEX{
   \Subsection{Functions on |BigNum|s}
   \Subsubsection{Allocation}
   \title{Allocation of a |BigNum|}
   \input{A length |nl|.}
   \output{An allocated |BigNum| of size |nl| or |NULL| if an error occured.}
   } 
   */
BigNum ZBNC(nl)
     BigNumLength nl; {}
#endif
#define ZBNC(l) (BigNum)ZENMalloc((l)*sizeof(BigNumDigit))

#if 0
/*
   \TEX{
   \title{Freeing a |BigNum|}
   \input{An allocated |BigNum|.}
   \side{|n| is freed.}
   } 
   */
void ZBNF(n)
     BigNum n; {}
#endif
#define ZBNF(n) ZENFree(n)


#if 0
/*
   \TEX{
   \title{True length of a |BigNum|}
   \input{An allocated |BigNum| and its length.}
   \output{The number of significant digits of |(n,nl)|.}
   \note{If |(n,nl)| is zero, returns 1.}
   } 
   */
BigNumLength ZBNNumDigits(n,nl)
     BigNum n;
     BigNumLength nl; {}
/*
  \TEX{
\Subsection{Assigning}

\title{Set a |BigNum| to zero}
   \input{An allocated |BigNum| and its length.}
   \side{The |BigNum| is filled with zeros.}
   } 
   */
void ZBNSetToZero(n,nl)
     BigNum n;
     BigNumLength nl; {}

/*
  \TEX{
  \title{Set a |BigNum| randomly}
   \input{An allocated |BigNum| and its length.}
   \side{The |BigNum| is filled randomly.}
   } 
   */
void ZBNSetRandom(n,nl)
     BigNum n;
     BigNumLength nl; {}
/*
  \TEX{
  \title{Assigning |BigNum|s}
   \input{Two allocated |BigNum|s and a length.}
   \side{The |BigNum| |m| is filled with |(n,nl)|.}
   \note{All kinds of overlapping are possible. No side effect if |nl| is zero.}
   } 
   */
void ZBNAssign(m,n,nl)
     BigNum m,n;
     BigNumLength nl; {}

/*
   \TEX{
\Subsection{Comparisons}

\title{Comparison of bignums}
   \input{Two allocated |BigNum|s and their lengths}
   \output{\begin{itemize}
   \item A positive integer if $|m| > |n|$;
   \item zero if $|m| = |n|$;
   \item A negative integer if $|m| < |n|$.\end{itemize}}
   } 
   */
int ZBNCompare(m,ml,n,nl)
     BigNum m,n;
     BigNumLength ml,nl;{}

/*
\title{Comparison of bignums}
   \input{Two allocated |BigNum|s and their true lengths.}
   \output{\begin{itemize}
   \item A positive integer if $|m| > |n|$;
   \item zero if $|m| = |n|$;
   \item A negative integer if $|m| < |n|$.\end{itemize}}
   \note{One must have $|m[ml-1]|\neq 0$ and $|n[nl-1]|\neq 0$.}
   } 
   */
int ZBNCompareLazy(m,ml,n,nl)
     BigNum m,n;
     BigNumLength ml,nl;{}

}
*/
/*
   \TEX{   
\Subsubsection{Tests}

\title{Test if two bignums are equal}
   \input{Two allocated |BigNum|s and their lengths}
   \output{The predicate |(n,nl) == (m,ml)|}
   } 
   */
int ZBNAreEqual(m, ml, n, nl)
     BigNum n,m;
     BigNumLength nl,ml; {}
#endif
#define ZBNAreEqual(mm, ml, nn, nl) !ZBNCompare(mm, ml, nn, nl)

#if 0
/*
   \TEX{
   \title{Test if a big integer value is equal to zero}
   \input{An allocated |BigNum| and its length.}
   \output{The predicate |(n,nl) == 0|}
   } 
   */
int ZBNIsZero(n,nl)
     BigNum n;
     BigNumLength nl; {}
/*
   \TEX{
   \title{Test if a big integer value is equal to one}
   \input{An allocated |BigNum| and its length.}
   \output{The predicate |(n,nl) == 1|}
   } 
   */
int ZBNIsOne(n,nl)
     BigNum n;
     BigNumLength nl; {}
#endif
# define ZBNIsOne(n, nl) ((ZBNNumDigits(n, (BigNumLength)(nl)) == 1 && ZBNDigit((n), 0) == (BigNumDigit)1) ? true : false)

#if 0
/*
   \TEX{
\Subsection{Arithmetic}
\Subsubsection{Addition}   

\title{Incrementation}
   \input{An allocated |BigNum| and its length, and a carry in that must be 0 or 1.}
   \output{The carry out.}
   \side{$|(n,nl)| += |c|$}
   \note{If |nl| is zero, nothing is done and the returned value is |c|.}
   } 
   */
BigNumDigit ZBNAddCarry(n,nl,c)
     BigNum n;
     BigNumLength nl;
     BigNumDigit c;{}
/*
   \TEX{
   \title{Addition}
   \input{Two allocated |BigNum|s, their lengths and a carry in that must be be 0 or 1.}
   \output{The carry out}
   \side{$|(n,nl)| += |(m,ml)| + |c|$.}
   \note{One must have $|ml| \le |nl|$.
   If $|ml|=0$, then |ZBNAdd(n,nl,m,0,c)| behaves like |ZBNAddCarry(n,nl,c)|.}
   } 
   */
BigNumDigit ZBNAdd(n,nl,m,ml,c)
     BigNum n,m;
     BigNumLength nl,ml;
     BigNumDigit c;{}


/*
   \TEX{
\Subsubsection{Subtraction}

\title{Additive inverse}
   \input{An allocated |BigNum|}
   \side{$|(n,nl)| \leftarrow 2^{|SIZE_BLOC|\times |nl|} -1 -|(n,nl)|$ }
   \note{If |nl| equals zero, does nothing.}
   } 
   */
void ZBNComplement(n,nl)
     BigNum n;
     BigNumLength nl; {}
/*
   \TEX{
   \title{Decrementation}
   \input{An allocated |BigNum|, its length and a borrow in that must be 0 or 1.}
   \output{The borrow out.}
   \side{$|(n,nl)| -= (1-|b|)$}
   \note{If |nl| is zero, nothing is done and the return value is |b|.}
   } 
   */
BigNumDigit ZBNSubtractBorrow(n,nl,b)
     BigNum n;
     BigNumLength nl;
     BigNumDigit b;{}

/*
   \TEX{
   \title{Subtraction}
   \input{Two allocated |BigNum|s, their length and a borrow in that must be be 0 or 1.}
   \output{The borrow out.}
   \side{$|(n,nl)| -= |(m,ml)| + (1-|b|)$}
   \note{One must have $|ml|\le |nl|$. If $|ml|=0$, then |ZBNSubtract(n,nl,m,0,b)| behaves like |ZBNSubtractBorrow(n,nl,b)|.}
   } 
   */
BigNumDigit ZBNSubtract(n,nl,m,ml,b)
     BigNum n,m;
     BigNumLength nl,ml;
     BigNumDigit b; {}

/*
   \TEX{
\Subsubsection{Shifting a |BigNum|}

\title{Shifting left}
   \input{An allocated |BigNum|, its length and a positive integer $0 \le |nb| < |SIZE_BLOC|$.}
   \output{The digit shifted out}
   \side{$|(n,nl)| <<= |nb|$}
   \note{If |nb| equals zero, nothing is done and the returned value is 0.}
   } 
   */
BigNumDigit ZBNShiftLeft(n,nl,nb)
     BigNum n;
     BigNumLength nl;
     unsigned int nb; {}     


/*
   \TEX{
\title{Shifting right}
   \input{An allocated  |BigNum|, its length and a positive integer $0 \le |nb| < |SIZE_BLOC|$.}
   \output{The digit shifted out}
   \side{$|(n,nl)| >>= |nb|$}
   \note{If |nb| equals zero, nothing is done and the returned value is 0.}
   } 
   */
BigNumDigit ZBNShiftRight(n,nl,nb)
     BigNum n;
     BigNumLength nl;
     unsigned int nb; {}     


/*
   \TEX{
   \input{shift}

\Subsubsection{Multiplication}

\title{Multiplication by a digit}
   \input{Two allocated |BigNum|s and a digit.}
   \output{A carry out.}
   \side{$|(n,nl)| += |(m,ml)| \times |d|$.}
   \note{One must have $|nl| \ge |ml|+1$.
   If |ml| equals zero, the return is 0 and nothing is done.
   }
   } 
   */
BigNumDigit ZBNMultiplyDigit(n,nl,m,ml,d)
     BigNum n,m;
     BigNumLength nl,ml;
     BigNumDigit d; {}

/*
   \TEX{
\title{Multiplication by a digit}
   \input{Two allocated |BigNum|s of length |nl| and a digit.}
   \output{The most significant digit of $|(m,ml)| \times |d|$.}
   \side{|(n,ml)| is filled with the |ml| least significant digits of $|(m,ml)| \times |d|$.}
   \note{One must have |n| distinct from |m|. }
   }
   */
BigNumDigit ZBNMultDigit(n,m,ml,d)
     BigNum n,m;
     BigNumLength ml;
     BigNumDigit d; {}

/*
   \TEX{
   \title{Multiplication}
   \input{Three allocated |BigNum|s and their lengths}
   \output{A carry out.}
   \side{$|(n,nl)| += |(a,al)| \times |(b,bl)|$.}
   \note{One must have $|nl| \ge |al|+|bl|$.
   If |bl| equals zero, there is no side effect and the carry out is 0.}
   } 
   */
BigNumDigit ZBNMultiply(n,nl,a,al,b,bl)
     BigNum n,a,b;
     BigNumLength nl,al,bl; {}
/*
   \TEX{
   \title{Multiplication}
   \input{Three allocated |BigNum|s and their lengths}
   \side{$|(n,nl)| \leftarrow |(a,al)| \times |(b,bl)|$.}
   \note{One must have |n| distinct from both |a| and |b|, $|nl| \ge |al|+|bl|$, and $|al| \ge |bl|$.}
   } 
   */
void ZBNMult(n,a,al,b,bl)
     BigNum n,a,b;
     BigNumLength al,bl; {}

/*
   \TEX{
\Subsubsection{Squaring}

\title{Squaring}
   \input{Two allocated |BigNum|s and their lengths}
   \output{A carry out.}
   \side{$|(n,nl)| += |(a,al)|^2$.}
   \note{One must have $|nl| \ge 2|al|$.
   If |al| equals zero, there is no side effect and the carry out is 0.}
   } 
   */
BigNumDigit ZBNSquare(n, nl, a, al)
     BigNum n,a;
     BigNumLength nl,al; {}
/*
   \TEX{
   \title{Squaring}
   \input{Two allocated |BigNum|s and a length}
   \side{$|(n,nl)| \leftarrow |(a,al)|^2 $.}
   \note{One must have |n| distinct from |a|, and $|nl| \ge 2|al|$.}
   } 
   */
void ZBNSqu(n,a,al)
     BigNum n,a;
     BigNumLength al; {}



/*
   \TEX{
\Subsubsection{Division}

\title{Division by a digit}
   \input{Two allocated |BigNum|s, a |BigNumLength| and a |BigNumDigit|.}
   \output{The remainder of the division $|n|/|d|$.}
   \side{The |BigNum| |(q,nl-1)| is filled with the quotient $\frac{|n|}{|d|}$.}
   \note{The value |d| must be greater than |ZBNDigit(n,nl-1)|.}
   } 
   */
BigNumDigit ZBNDivideDigit(q, n, nl, d)
     BigNum q,n;
     BigNumLength nl;
     BigNumDigit d; {}


/*
   \TEX{
   \title{Division}
   \input{Two allocated |BigNum|s and their lengths.}
   \side{The quotient and the remainder of $|(n,nl)|/|(d,dl)|$ are stored in |n|. The |dl| least significant digits contains the remainder, whereas the $|nl|-|dl|$ most significant digits contains the quotient.}
   \note{The value |ZBNDigit(d,dl-1)| must be strictly greater than |ZBNDigit(n,nl-1)|.}
   } 
   */
void ZBNDivide(n, nl, d, dl)
     BigNum n,d;
     BigNumLength nl,dl; {}

/*
   \TEX{
\Subsubsection{Miscellaneous operations}
These operations are preliminary. They may change in future versions.
   \input{log}
   \input{sqrt}
   
\Subsection{Input/output}
\input{io}
   } 
   */
#endif




#define ZBNPrint(n, nl) ZBNPrintToFile(stdout, n, (BigNumLength)(nl), 10)
#if 0
/*
   \TEX{
   \Subsection{Greatest Common Divisor}
   \input{gcd}
   
   \Subsection{Modular operations}
   \title{Reduction of a |BigNum| modulo another}
   \input{Two allocated |BigNum|s and their lengths, and a pointer on a non allocated |BigNum|.}
   \output{|ZENERR| if an error occured, the length of the result otherwise.}
   \side{|*p_m| is allocated and set to $|n| \bmod |m| $.}
   } 
   */
BigNumLength ZBNModulo(p_m,n,nl,m,ml)
     BigNum *p_m,n,m;
     BigNumLength nl,ml;{}
#endif
#define ZBNModulo(p_m,n,nl,m,ml) ZBNModuloLazy(p_m,n,ZBNNumDigits(n,nl),m,ZBNNumDigits(m,ml))

#if 0
/*
   \TEX{
   \input{modulo}
   
   \Subsection{Hamming weight of a |BigNum|}
   \title{Number of bits of a digit}
   \input{A |BigNumDigit|}
   \output{The number of non zero bits in |d|}
   \note{This results in a macro using tabulated datas, or a function call if |LOW_MEMORY| is enabled. Use the function |ZBNWeightDigit| if you want to use side effects in |d|.}
   } 
   */
unsigned int ZBNWeightDigitFast(d)
     BigNumDigit d; {}
#endif

#define _ZBNWeightDigitFast(D) (sizeof(BigNumDigit) == sizeof(short)?__ZEN_WEIGHT[D]:(sizeof(BigNumDigit) == sizeof(short)<<1 ? __ZEN_WEIGHT[(D) & 65535]+__ZEN_WEIGHT[((D)>>16)] : __ZEN_WEIGHT[(D) & 65535]+__ZEN_WEIGHT[((D)>>48)]+__ZEN_WEIGHT[((D) >> 16)& 65535]+__ZEN_WEIGHT[((D)>>32)& 65535] ))
#if LOW_MEMORY
# define ZBNWeightDigitFast(D) ZBNWeightDigit(D)
#else
extern unsigned int __ZEN_WEIGHT[];
# define ZBNWeightDigitFast(D) _ZBNWeightDigitFast(D)
#endif
/*
   \TEX{
   \input{weight}   
   } 
   */
extern int ZBNCompareLazy ___((BigNum, BigNumLength, BigNum, BigNumLength));

extern unsigned int ZBNWeightDigit ___((BigNumDigit));
extern int ZBNWeight ___((BigNum,BigNumLength));

extern char* ZBNPrintToString ___((BigNum,BigNumLength,int));
extern int ZBNReadFromString ___((BigNum *,BigNumLength *, char *, int));
extern int ZBNReadFromStringLazy ___((BigNum ,BigNumLength , char *, int));

extern char* ZBNPutToString ___((BigNum,BigNumLength));
extern char* ZBNPutToStringLazy ___((BigNum,BigNumLength));
extern int ZBNGetFromString ___((BigNum *,BigNumLength *, char *));
extern int ZBNGetFromStringLazy ___((BigNum ,BigNumLength , char *));

extern int ZBNPrintToFile ___((FILE *,BigNum,BigNumLength,int));
extern int ZBNReadFromFile ___((BigNum *,BigNumLength *,FILE *,int));
extern int ZBNReadFromFileLazy ___((BigNum ,BigNumLength ,FILE *,int));

extern int ZBNPutToFile ___((FILE *,BigNum,BigNumLength));
extern int ZBNPutToFileLazy ___((FILE *,BigNum,BigNumLength));
extern int ZBNGetFromFile ___((BigNum *, BigNumLength *,FILE *));
extern int ZBNGetFromFileLazy ___((BigNum , BigNumLength ,FILE *));

extern double ZBNLog ___((BigNum,BigNumLength));
extern BigNumLength ZBNSqrt ___((BigNum,BigNum,BigNumLength));

extern BigNumLength ZBNModuloLazy ___((BigNum *,BigNum,BigNumLength,BigNum,BigNumLength));

extern void ZBNModAdd ___((BigNum, BigNumLength, BigNum, BigNumLength, BigNum, BigNumLength));
extern void ZBNModNegate ___((BigNum, BigNumLength, BigNum, BigNumLength, BigNum, BigNumLength));
extern void ZBNModSubtract ___((BigNum, BigNumLength, BigNum, BigNumLength, BigNum, BigNumLength));
extern void ZBNModMultiply ___((BigNum, BigNumLength, BigNum, BigNumLength, BigNum, BigNumLength, BigNum, BigNumLength));
extern void ZBNModMultiplyKaratsubaBuffer ___((BigNum, BigNumLength, BigNum, BigNumLength, BigNum, BigNumLength, BigNum, BigNum, BigNumLength));
extern void ZBNModSquare ___((BigNum, BigNumLength, BigNum, BigNumLength, BigNum, BigNumLength));
extern void ZBNModSquareKaratsubaBuffer ___((BigNum, BigNumLength, BigNum, BigNumLength, BigNum, BigNum, BigNumLength));
extern void ZBNModLongSquare ___((BigNum, BigNumLength, BigNum, BigNumLength, BigNum, BigNumLength));
extern void ZBNModSmallSquare ___((BigNum, BigNumLength, BigNum, BigNumLength, BigNum, BigNumLength));

extern void ZBNAnyShiftRight ___((BigNum,BigNumLength,unsigned int));
extern void ZBNAnyShiftLeft ___((BigNum,BigNumLength *,BigNum,BigNumLength, unsigned int));
#endif
