/*-*-LaTeX-*-
  This is a documentation file of the ZEN library.
  There is no C code in this file, only LaTeX commands.


\TEX{

  Except the original way we process the |LaTeX|~\cite{latex}
  documentation, |Makefiles| generation with |imake|~\cite{imake} and version
  control of sources with |CVS|~\cite{cvs} inherit in a large part from
  the programming environment work done by the |LiPS| project
  (\cite{lips}).
  
  Once explained in section \ref{parser} how we interlace
  documentation and C code in our sources, we describe how we create
  |Makefiles| in section \ref{makefiles}.

  \Section{The documentation parser}
\label{parser}
  \input{parser}

  \Section{Makefiles}
  \label{makefiles}

  All this can be found as it in \cite{imake}.

  \Subsection{Makefiles are non portable}
  
  In the UNIX world, software development typically involves |make|,
  using a |Makefile| describing how to build and install the programs
  in which we're interested. In essence, |make| functions as a
  command-generating engine, and the |Makefile| controls which command
  to generate for specific targets. This is a tremendous convenience
  for the programmer, since it eliminates the need to type out a bunch
  of commands each time you want to build something.

  Unfortunately, |Makefiles| are not portable. The commands that make
  so conveniently generates at the drop of a hat are subject to
  variation from system to system, so a |Makefile| that works on my
  system may not work on yours. Ideally, we'd be able to take a
  software project, put it on another machine, compile and install
  it, and it would run. But it doesn't always work that way.

  Suppose we're writing a small C program, |myname| consisting of a
  single source file |myname.c|:

  }
\CiteC
*/
#include <stdio.h>

int main()
{
  char *getenv();
  char *p;

  p = getenv ("USER");
  if (p == (char *)NULL) {
    printf("Cannot tell your user name\n");
  }
  else {
    printf("Your user name is %s\n", p);
  }
  exit(0);
}

/*
\EndCiteC
\TEX{
  The |Makefile| for the program might look like this:
  }
\CiteC
*/
myname: myname.o
        cc -o myname myname.o
install: myname
        install myname /usr/local
/*
\EndCiteC
\TEX{
  To build and install myname, we'd use these commands:
  }
\CiteC
*/
% make myname
cc  -c myname.c
cc -o myname myname.o
% make install
install myname /usr/local
/*
\EndCiteC
\TEX{
  
  All our friends are astonished at the use fullness of this program
  and immediately begin asking us for the source
  code. We give it to one of them, together with the |Makefile|, and
  he lopes off to install it on his machine. The next day he sends
  back the following report:
  \begin{itemize}
  \item We don't have any install program, so I used |cp|.
  \item Our local installation directory is |/usr/local/bin| rather
  than |/usr/local|.
  \end{itemize}

  He sends back a revised |Makefile|, having made the necessary
  changes, and also having conveniently parameterized the things he
  found to be different on his system, using the |make| variables
  |INSTALL| and |BINDIR| at the beginning of the |Makefile|:
  }
\CiteC
*/
INSTALL = cp
BINDIR = /usr/local/bin
myname: myname.o
        cc -o myname myname.o
install: myname
        install myname /usr/local
/*
\EndCiteC
\TEX{
  This is a step forward. The installation program and directory as
  specified in this |Makefile| are no longer correct for our machine,
  but now they're parameterized and thus easily located and
  modified. We change the values back to |install| and
  |/usr/local/bin/| so they'll work on our machine, and both we and
  our friend are happy.

  Alas, our idyllic state of mind doesn't last long. We give the
  program to another friend. She builds it and reports back her
  changes:
  \begin{itemize}
  \item I prefer to use |gcc|, not |cc|.
  \item Our C library is broken and doesn't contain |getenv()|, I have
  to link in |-lc_aux| as well.
  \item The environment variable |USER| isn't used on our system, but
  |LOGNAME| is, so I used that instead.
  \end{itemize}

  She, too, sends back a revised |Makefile|, further parameterized:
}
\CiteC
*/
CC = gcc
EXTRA_LIBS = -lc_aux
CFLAGS = -DUSE_LOGNAME
INSTALL = cp
BINDIR = /usr/local/bin
myname: myname.o
        $(CC) $(CFALGS) $(EXTRA_LIBS) -o myname myname.o
install: myname
        install myname /usr/local
/*
\EndCiteC
$
\TEX{
  and a revised file |myname.c|:
}
\CiteC
*/
#include <stdio.h>

int main()
{
  char *getenv();
  char *p;

#ifdef USE_LOGNAME
  p = getenv ("LOGNAME");
#else
  p = getenv ("USER");
#endif
  if (p == (char *)NULL) {
    printf("Cannot tell your user name\n");
  }
  else {
    printf("Your user name is %s\n", p);
  }
  exit(0);
}
/*
\EndCiteC
\TEX{

  Now, with a simple change to the |Makefile|, |myname| can be
  recompiled on machines that use either environment variable. This is
  encouraging. Using our vast and ever increasing porting experience,
  |myname| now works on more machines-three ! And the revised
  |Makefile| is better than the original:
  \begin{itemize}
  \item It parameterizes all the non portability we've encountered so
  far.
  \item Nonportabilities are identified explicitly at the beginning of
  the |Makefile|.
  \item They're easily modified by editing |Makefile|.
  \end{itemize}

  However, we begin to notice uneasily that the number of things that
  might need changing from machine to machine is increasing. They were
  only two differences between machines when we built |myname| on two
  systems. Now we can build it on three systems but the number of
  differences has increased to five. How many differences will we find
  when we attempt to port |myname| to additional systems ? As the
  |Makefile| editing job gets bigger, it becomes more difficult. Each
  nonportability adds another increment to the burden incurred each time
  |myname| is built on a different machine.

  Also, we'd better keep a list of parameters appropriate for each
  type of machine on which the program is known to run, or we'll
  forget them. And we need to distribute the list to other people who
  want to build the program, so they can consult it to see how they
  might need to modify the makefile on their |systems|.

  The previous example outlines mainly four drawbacks of |make|:
  \begin{itemize}
  \item There are no conditionals:\\ If we could assign values to
    |make| variables conditionally, we could parameterize machine
    dependencies by assigning variable values appropriate to a given
    machine type.  To some extend, we can use the shell's
    |if|-statement syntax to get around lack of conditionals in target
    entries, but it's clunky to do so.

  \item There is no flow control:\\
    Loops and iterators must be done using shell commands to simulate
    them.

  \item It's difficult to make global changes:\\ For example, suppose
    you use BINDIR to name your installation directory. You can assign
    it the proper value in each |Makefile| in your project. But if you
    decide to change it, you must find and change every |Makefile|.
    ALternatively, you can assign |BINDIR|, once in the top level
    |Makefile| and pass the value to |make| commands in subdirectories
    using suitable recursive rules. But, then you can't run |make|
    directly in a subdirectory because |BINDIR| won't have the proper
    value. Recursive rules are difficult to write correctly, anyway.

    \item Header file dependencies for C source files are inherently
    nonportable:\\
      Different systems organize header files differently. You might
      even have multiply sets on systems that support development
      under more than one environment (BSD, System V, \ldots). This
      makes it impossible to list header file dependencies statically
      in the |Makefile|. They must be computed on the target machine
      at build time.
  \end{itemize}

  \Subsection{The solution |imake|}

  With |imake|, you don't write a |Makefile|, instead you write an
  |imakefile|, a machine independent description of the targets you
  want to build. Machine dependencies are centralized into a set of
  configuration files isolated in their own directory, |config|, so
  they don't appear in the |imakefile|. |imake| merges information
  obtained from |config| and |imakefile| to generates a properly
  configured |Makefile|. Then you run |make| to build your programs.

  From a practical point of few, |imake| is just an interface for the
  preprocessor |cpp|. In a first approximation, the |Makefile| that
  |imake| generates, would be more or less the file obtained with
  |cpp -I. -Iconfig config/Imake.tmpl Makefile|. |cpp| usually defines
  specific variables for each machine and these variables are tested in
  |Imake.templ| to include the corresponding files of |config|. Then
  |Imake.templ| includes |host.def| (which includes |dir.def|) and
  finally the current |imakefile|.

  As |cpp| parses them, each |imakefile| is a collection of
  macros. Most of them are defined by default in the file
  |config/imake.rules|.

}
*/
